#
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Accelerate training of neural networks using importance sampling."""

__author__ = "Angelos Katharopoulos"
__copyright__ = "Copyright (c) 2017 Idiap Research Institute"
__license__ = "MIT"
__maintainer__ = "Angelos Katharopoulos"
__email__ = "angelos.katharopoulos@idiap.ch"
__url__ = "http://www.idiap.ch/~katharas/importance-sampling/"
__version__ = "0.2"

from training import ImportanceTraining, ApproximateImportanceTraining
